from evo_science.entities.metrics.base_metric import BaseMetric


class Slope(BaseMetric):
    name = "Slope"

    def _calculate_np(self, y_true, y_pred):
        return self.model.model.coef_[0][0]



"""CLI interface for evo_science project.

Be creative! do whatever you want!

- Install click or typer and create a CLI app
- Use builtin argparse
- Start a web application
- Import things from your .base module
"""


def main():  # pragma: no cover
    """
    The main function executes on commands:
    `python -m evo_science` and `$ evo_science `.

    This is your program's entry point.

    You can change this function to do whatever you want.
    Examples:
        * Run a test suite
        * Run a server
        * Do some other stuff
        * Run a command line application (Click, Typer, ArgParse)
        * List all available tasks
        * Run an application (Flask, FastAPI, Django, etc.)
    """
    print("This will do something")



class MetricLib:
    metrics = {
        "RMSE": "from src.entities.metrics.rmse import RMSE",
        "Slope": "from evo_science.entities.metrics.slope import Slope",
        "MAE": "from src.entities.metrics.mae import MAE",
    }

    def __init__(self):

        raise NotImplementedError("This class is not implemented yet")

    @classmethod
    def get_metric(cls, name):
        if name not in cls.metrics:
            raise ValueError(f"Metric {name} is not supported")
        exec(cls.metrics[name])



from .base_metric import BaseMetric
from .slope import Slope
from .rmse import RMSE
from .mae import MAE
from .error_ave import ErrorAve
from .r_squared import RSquared
from .error_ave import ErrorAve
from .error_std import ErrorStd



from src.entities.metrics.base_metric import BaseMetric


class Intercept(BaseMetric):
    name = "Intercept"

    def _calculate_np(self, y_true, y_pred):
        return self.model.model.intercept_[0]



import pandas as pd
import xgboost
from dtreeviz import dtreeviz
from matplotlib import pyplot as plt

from evo_science.entities.features import FeatureSet


class BaseModel:
    pass


class XGBoost(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.model = xgboost.XGBRegressor(**kwargs)

    def fit(self, x, y):
        self.model.fit(x.to_numpy(), y.to_numpy())

    def plot(self, x: FeatureSet, y: FeatureSet):
        raise NotImplementedError("XGBoost does not support plotting")

    def _calculate_coefficients(self, x: FeatureSet):
        coef = self.model.feature_importances_
        feature_names = x.column_names
        coef_dict = dict(zip(feature_names, coef))
        df = pd.DataFrame(coef_dict.items(), columns=["Feature", "Coefficient"])

        return df



from evo_science.entities.models.base_model import BaseModel


class DecisionTree(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.decision_tree = None
        raise NotImplementedError("This class is not implemented yet.")

    def fit(self, x, y):
        pass

    def predict(self, x):
        pass

    def evaluate(self, x, y):
        pass

    def plot(self, x, y):
        pass



import logging


class EvoScienceLogger(logging.Logger):
    def __init__(self, name="CPA Logger", level=logging.NOTSET):
        super().__init__(name, level)
        self.setLevel(logging.DEBUG)
        self.addHandler(logging.StreamHandler())



import typing

import numpy as np
import pandas as pd
from tabulate import tabulate

from evo_science import BaseFeature
from evo_science.entities.cpa_logger import EvoScienceLogger


class FeatureSet:
    def __init__(
        self,
        features: typing.List["BaseFeature | type"],
        auto_skip_nan: bool = True,
        parents=None,
    ):
        self.auto_skip_nan = auto_skip_nan
        for i, feature in enumerate(features):
            if isinstance(feature, type):
                features[i] = feature()

        self._features = features
        self.logger = EvoScienceLogger()
        self._is_built = False
        self.__parents = parents
        if not parents:  # if no parents, then it is the root
            table = tabulate(
                [[f.column_name, f.column_type, f.categories] for f in features],
                headers=["Feature", "Type", "Categories"],
                tablefmt="grid",
            )
            self.logger.info(
                f"FeatureSet is built with the following features:\n{table}"
            )

    @property
    def column_names(self):
        return [x.column_name for x in self.features]

    @property
    def hard_column_names(self):
        column_names = []
        for feature in self.features:
            if feature.is_skip:
                continue
            if feature.virtual:
                column_names += feature.use_columns
            else:
                column_names.append(feature.column_name)

    @property
    def features(self):
        return [x for x in self._features if not x.is_skip]

    def clean(self, df):
        for feature in self.features:
            df = feature.clean(df=df)
        return df

    def encode(self, df):
        for feature in self.features:
            if feature.is_skip:
                continue
            df.loc[:, feature.column_name] = feature.encode(
                df[feature.column_name].values
            )
        return df

    def build(self, csv_path=None, df=None, train_test_split=None, seed=42):
        assert csv_path or df, "Either csv_path or df should be provided."
        assert not (csv_path and df), "Only one of csv_path or df should be provided."

        df = pd.read_csv(csv_path) if csv_path else df

        df = self.build_virtual_feature(df)  # for virtual features

        if self.auto_skip_nan:
            for i, feature in enumerate(self.features):
                try:
                    df_feature = df[feature.column_name]
                except KeyError as e:
                    self.logger.error(
                        f"Feature {feature.column_name} is not found in the data. {e}\n valid columns are {df.columns}"
                    )
                    raise e
                is_all_nan = np.all(pd.isna(df_feature))
                if is_all_nan:
                    self.features[i].is_skip = True
                    self.logger.warning(
                        f"Feature {feature.column_name} is skipped because it contains only NaN values."
                    )

        df = self.clean(df)
        df = self.fillna(df)
        df = self.encode(df)

        if train_test_split:
            train_size = int(len(df) * (1 - train_test_split))
            df_train = df.sample(n=train_size, random_state=seed)
            df_test = df.drop(df_train.index)
            for i, feature in enumerate(self.features):
                self.features[i].train_data = df_train[feature.column_name].values
                self.features[i].test_data = df_test[feature.column_name].values
            return df_train, df_test
        else:
            for i, feature in enumerate(self.features):
                self.features[i].train_data = df[feature.column_name].values
                self.features[i].test_data = df[feature.column_name].values
        self._is_built = True
        if self.__parents:
            for parent in self.__parents:
                parent._is_built = True
        return df

    def __add__(self, other):
        return FeatureSet(self.features + other.features, parents=[self, other])

    def to_numpy(self, is_train=True):
        # Adding dtype=object, for handling inhomogeneous data
        data = [
            feature.to_numpy(is_train=is_train)
            for feature in self.features
            if not feature.is_skip
        ]
        data = np.array(data).T
        return data

    def fillna(self, df):
        for feature in self.features:
            df[feature.column_name] = feature.fillna(df[feature.column_name])
        return df

    @property
    def df(self):
        data = self.to_numpy()
        column_names = [x.column_name for x in self.features if not x.is_skip]
        df = pd.DataFrame(data=data, columns=column_names)
        return df

    def build_virtual_feature(self, df):
        for feature in self.features:
            if feature.virtual:
                df[feature.column_name] = feature.compute(df)
        return df



import matplotlib.pyplot as plt
import pandas as pd
from sklearn.linear_model import LinearRegression

from evo_science import FeatureSet, BaseModel


class LinearRegressionModel(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.model = LinearRegression(**kwargs)

    def plot(self, x: FeatureSet, y: FeatureSet):
        x_data = x.to_numpy()
        y_data = y.to_numpy()
        y_pred = self.predict(x_data)
        dimension = x_data.shape[1]
        if dimension == 1:
            plt.scatter(x_data, y_data, color="blue")
            plt.plot(x_data, y_pred, color="green", linewidth=3)
            plt.show()
        else:
            for i in range(dimension):
                x_dim = x_data[:, i]
                x_dim = x_dim.reshape(-1, 1)
                plt.scatter(x_dim, y_data, color="blue")
                plt.scatter(x_dim, y_pred, color="red")
                # prepare line function y=mx+c
                m = self.model.coef_[0][i]
                c = self.model.intercept_[0]
                y_line = m * x_dim + c
                plt.plot(x_dim, y_line, color="green", linewidth=3)

                plt.xlabel(x.column_names[i])
                plt.show()

    def _calculate_coefficients(self, x: FeatureSet):
        coef = self.model.coef_[0]
        feature_names = x.column_names
        coef_dict = dict(zip(feature_names, coef))
        df = pd.DataFrame(coef_dict.items(), columns=["Feature", "Coefficient"])
        return df



import numpy as np

import typing

if typing.TYPE_CHECKING:
    from evo_science import BaseModel, FeatureSet, BaseFeature


class BaseMetric:
    name = "base_metric"

    def __init__(
        self,
        model: "BaseModel",
        feature_set: "FeatureSet",
        target_feature: "BaseFeature",
    ):
        self.feature_set = feature_set
        self.target_feature = target_feature
        self.model = model

    def _calculate_np(self, y_true: np.array, y_pred: np.array):
        """
        Calculate the metric using numpy arrays.
        Args:
            y_true (np.array): True target values.
            y_pred (np.array): Predicted target values.
        Returns:
            float: Metric value.
        """
        raise NotImplementedError

    def calculate(self):
        """
        Calculate the slope of the model predictions.

        Returns:
            float: Slope of the model predictions.
        """
        y_true_np = self.target_feature.to_numpy(is_train=False)
        y_pred_np = self.model.predict(self.feature_set.to_numpy(is_train=False))

        return self._calculate_np(y_true_np, y_pred_np)



import numpy as np


class BaseFeature:
    virtual = False
    categories = None
    column_type = None
    column_name = None
    use_columns = None

    def __init__(self):
        self.train_data = None
        self.test_data = None
        self.is_skip = False
        self.column_name = self.column_name or self.__class__.__name__

    @property
    def data(self):
        """Return the data for the feature. If test data is available, return the concatenated train and test data."""
        data = self.train_data
        if self.test_data is not None:
            data = np.concatenate((data, self.test_data), axis=0)
        return data

    def clean(self, df):
        return df

    def encode(self, x: np.array) -> np.array:
        if self.column_type == "category":
            return self.encode_category(x)
        return x.astype(self.column_type)

    def build(self, df):
        if self.virtual:
            df[self.column_name] = self.compute(df)

        df = self.clean(df)
        df[self.column_name] = self.fillna(df[self.column_name])
        df[self.column_name] = self.encode(df[self.column_name].values)

        return df

    def to_numpy(self, is_train=True):
        if is_train:
            return self.train_data
        else:
            return self.test_data

    def is_all_nan(self):
        """
        Check if the feature contains only NaN values.
        """
        return np.all(np.isnan(self.train_data))

    def encode_category(self, x: np.array):
        if isinstance(self.categories, list):
            return [self.categories[_x] for _x in x]
        if isinstance(self.categories, dict):
            return [self.categories[_x] for _x in x]
        if self.categories == "__auto__":
            all_categories = set(x)
            # build auto encoding
            all_categories = list(all_categories)
            all_categories.sort()
            np_x = np.zeros(len(x))
            for i, category in enumerate(all_categories):
                np_x[x == category] = i
            return np_x
        if self.categories == "__one_hot__":
            all_categories = x.unique()
            # build one hot encoding
            np_x = np.zeros((len(x), len(all_categories)))
            for i, category in enumerate(all_categories):
                np_x[:, i] = x == category
            return np_x
        raise ValueError(f"Unknown category type: {self.categories}")

    def fillna(self, x):
        is_na_values = x.isna()
        if is_na_values.any():
            raise ValueError(f"Feature {self.column_name} contains NaN values.")
        return x

    def compute(self, df):
        raise NotImplementedError("Virtual feature must implement compute method")



from sklearn.linear_model import LogisticRegression

from evo_science.entities.features import FeatureSet
from evo_science.entities.models.base_model import BaseModel


class LogisticRegressionModel(BaseModel):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.model = LogisticRegression(**kwargs)

    def fit(self, x: FeatureSet, y: FeatureSet):
        x_data = x.to_numpy()
        y_data = y.to_numpy().flatten()
        self.model.fit(x_data, y_data)



import unittest
import sys

from examples.titanic_survival_prediction.linear_regression_model import (
    example_lr_model,
)

# appending the needed path for import
sys.path.append("/evo_science")
sys.path.append("/examples/titanic-survival-prediction")


class TestExample(unittest.TestCase):
    def test_execute_lr_example(self):
        # set working directory
        example_lr_model()



from evo_science.entities.metrics.base_metric import BaseMetric


class ErrorAve(BaseMetric):
    name = "Error Ave"

    def _calculate_np(self, y_true, y_pred):
        return (y_true - y_pred).mean()



import typing

import pandas as pd
from tabulate import tabulate

from evo_science import FeatureSet
from evo_science.entities.metrics import BaseMetric
from evo_science.metric_lib import MetricLib


class BaseModel:

    def __init__(self, **kwargs):
        self.model = None

    def fit(self, x: FeatureSet, y: FeatureSet):
        """
        Fit the model to the data.

        Args:
            x (FeatureSet): Feature set for training data.
            y (FeatureSet): Target set for training data.
        """
        assert x._is_built and y._is_built, "FeatureSet is not built yet."
        x_data = x.to_numpy(is_train=True)
        y_data = y.to_numpy(is_train=True)
        self.model.fit(x_data, y_data)

    def predict(self, X):
        """
        Predict using the fitted model.

        Args:
            X (array-like): Features data for prediction.

        Returns:
            array-like: Predicted target values.
        """
        return self.model.predict(X)

    def evaluate(
        self,
        x: FeatureSet,
        y: FeatureSet,
        size: float = None,
        metrics: typing.List["str | BaseMetric | type"] = None,
        **kwargs,
    ):
        """
        Evaluate the model performance.

        Args:
            x (FeatureSet): Feature set for test data.
            y (FeatureSet): Target set for test data.
            size (float, optional): Size of the data to be evaluated. The fraction of the total data used.

        Returns:
            float: Mean Squared Error of the model predictions.
        """
        table_records = []
        for i, metric in enumerate(metrics):
            if isinstance(metric, str):
                metric: type = MetricLib.get_metric(metric)
                metrics[i] = metric(model=self, feature_set=x, target_feature=y)
            elif isinstance(metric, type):
                metrics[i] = metric(model=self, feature_set=x, target_feature=y)
            else:
                assert issubclass(
                    metric.__class__, BaseMetric
                ), f"{metric.__class__.__name__} is not a subclass of BaseMetric"

        for metric in metrics:
            metric_value = metric.calculate()
            table_records.append([metric.name, metric_value])

        table = tabulate(
            tabular_data=table_records,
            headers=["Metric", "Value"],
            tablefmt="orgtbl",
        )
        print(table)

    def calculate_coefficients(self, x: FeatureSet) -> pd.DataFrame:
        """
        Get the coefficients of the features x to the target y.
        """
        df_cof = self._calculate_coefficients(x)

        table = tabulate(
            tabular_data=df_cof.values, headers=df_cof.columns, tablefmt="orgtbl"
        )
        print(table)
        return df_cof

    def plot(self, x: FeatureSet, y: FeatureSet, **kwargs):
        """
        Visualize the model predictions.

        Args:
            x (FeatureSet): Feature set for test data.
            y (FeatureSet): Target set for test data.
        """
        raise NotImplementedError

    def _calculate_coefficients(self, x):
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement _calculate_coefficients method."
        )



from evo_science.entities.features.base_feature import BaseFeature
from evo_science.entities.features.feature_set import FeatureSet
from evo_science.entities.features.feature_types import FEATURE_TYPES
from evo_science.entities.models.base_model import BaseModel



from . import BaseMetric


class RSquared(BaseMetric):
    name = "R Squared"

    def _calculate_np(self, y_true, y_pred):
        ss_res = ((y_true - y_pred) ** 2).sum()
        ss_tot = ((y_true - y_true.mean()) ** 2).sum()

        return 1 - (ss_res / ss_tot)
